@@ -1,8 +1,8 @@
 /*
- * MGA-1064, MGA-G100, MGA-G200, MGA-G400 RAMDAC driver
+ * MGA-1064, MGA-G100, MGA-G200, MGA-G400, MGA-G550 RAMDAC driver
  */
 
-/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/mga/mga_dacG.c,v 1.40 2000/10/24 22:45:07 dawes Exp $ */
+/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/mga/mga_dacG.c,v 1.49 2002/01/11 15:42:57 dawes Exp $ */
 
 /*
  * This is a first cut at a non-accelerated version to work with the
@@ -49,33 +49,6 @@
 
 #define OPTION1_MASK_PRIMARY	0xFFFC0FF
 
-/*
- * Read/write to the DAC via MMIO 
- */
-
-/*
- * These were functions.  Use macros instead to avoid the need to
- * pass pMga to them.
- */
-
-#define inMGAdreg(reg) INREG8(RAMDAC_OFFSET + (reg))
-
-#define outMGAdreg(reg, val) OUTREG8(RAMDAC_OFFSET + (reg), val)
-
-#define inMGAdac(reg) \
-	(outMGAdreg(MGA1064_INDEX, reg), inMGAdreg(MGA1064_DATA))
-
-#define outMGAdac(reg, val) \
-	(outMGAdreg(MGA1064_INDEX, reg), outMGAdreg(MGA1064_DATA, val))
-
-#define outMGAdacmsk(reg, mask, val) \
-	do { /* note: mask and reg may get evaluated twice */ \
-	    unsigned char tmp = (mask) ? (inMGAdac(reg) & (mask)) : 0; \
-	    outMGAdreg(MGA1064_INDEX, reg); \
-	    outMGAdreg(MGA1064_DATA, tmp | (val)); \
-	} while (0)
-
-
 static void MGAGRamdacInit(ScrnInfoPtr);
 static void MGAGSave(ScrnInfoPtr, vgaRegPtr, MGARegPtr, Bool);
 static void MGAGRestore(ScrnInfoPtr, vgaRegPtr, MGARegPtr, Bool);
@@ -137,6 +110,7 @@ MGAGCalcClock ( ScrnInfoPtr pScrn, long f_out,
 		post_div_max = 7;
 		break;
 	case PCI_CHIP_MGAG400:
+	case PCI_CHIP_MGAG550:
 		ref_freq     = 27050.5;
 		feed_div_min = 7;
 		feed_div_max = 127;
@@ -237,6 +211,12 @@ MGAGSetPCLK( ScrnInfoPtr pScrn, long f_out )
 	/* The actual frequency output by the clock */
 	double f_pll;
 
+	if(MGAISGx50(pMga)) {
+		MGAG450SetPLLFreq(pScrn, f_out);
+                pReg->PIXPLLCSaved = FALSE;
+		return;
+	}
+
 	/* Do the calculations for m, n, p and s */
 	f_pll = MGAGCalcClock( pScrn, f_out, &m, &n, &p, &s );
 
@@ -275,7 +255,21 @@ MGAGInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	MGARegPtr pReg;
 	vgaRegPtr pVga;
 	MGAFBLayout *pLayout;
-	
+	xMODEINFO ModeInfo;
+
+	ModeInfo.ulDispWidth = mode->HDisplay;
+        ModeInfo.ulDispHeight = mode->VDisplay;
+        ModeInfo.ulFBPitch = mode->HDisplay; 
+        ModeInfo.ulBpp = pScrn->bitsPerPixel;    
+        ModeInfo.flSignalMode = 0;
+        ModeInfo.ulPixClock = mode->Clock;
+        ModeInfo.ulHFPorch = mode->HSyncStart - mode->HDisplay;
+        ModeInfo.ulHSync = mode->HSyncEnd - mode->HSyncStart;
+        ModeInfo.ulHBPorch = mode->HTotal - mode->HSyncEnd;
+        ModeInfo.ulVFPorch = mode->VSyncStart - mode->VDisplay;
+        ModeInfo.ulVSync = mode->VSyncEnd - mode->VSyncStart;
+        ModeInfo.ulVBPorch = mode->VTotal - mode->VSyncEnd;
+
 	pMga = MGAPTR(pScrn);
 	pReg = &pMga->ModeReg;
 	pVga = &VGAHWPTR(pScrn)->ModeReg;
@@ -335,9 +329,13 @@ MGAGInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 		pReg->Option2 = 0x0000007;
 		break;
 	case PCI_CHIP_MGAG400:
+	case PCI_CHIP_MGAG550:
 #ifdef USEMGAHAL
 	       MGA_HAL(break;);
 #endif
+	       if (MGAISGx50(pMga))
+		       break;
+
 	       if(pMga->Dac.maxPixelClock == 360000) {  /* G400 MAX */
 	           if(pMga->OverclockMem) {
 			/* 150/200  */
@@ -379,7 +377,7 @@ MGAGInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	case PCI_CHIP_MGAG200_PCI:
 	default:
 #ifdef USEMGAHAL
-	       MGA_HAL(break;);
+		MGA_HAL(break;);
 #endif
 		if(pMga->OverclockMem) {
                      /* 143 Mhz */
@@ -528,11 +526,30 @@ MGAGInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
 	if (mode->Flags & V_DBLSCAN)
 		pVga->CRTC[9] |= 0x80;
 
+	if(MGAISGx50(pMga)) {
+		OUTREG(MGAREG_ZORG, 0);
+	}
+
 	MGAGSetPCLK(pScrn, mode->Clock);
 	);	/* MGA_NOT_HAL */
 
 	/* This disables the VGA memory aperture */
 	pVga->MiscOutReg &= ~0x02;
+
+	/* Urgh. Why do we define our own xMODEINFO structure instead 
+	 * of just passing the blinkin' DisplayModePtr? If we're going to
+	 * just cut'n'paste routines from the HALlib, it would be better
+	 * just to strip the MacroVision stuff out of the HALlib and release
+	 * that, surely?
+	 */
+        /*********************  Second Crtc programming **************/
+        /* Writing values to crtc2[] array */
+        if (pMga->SecondCrtc)
+        {
+            MGACRTC2Get(pScrn, &ModeInfo); 
+            MGACRTC2GetPitch(pScrn, &ModeInfo); 
+            MGACRTC2GetDisplayStart(pScrn, &ModeInfo,0,0,0);
+        }
 	return(TRUE);
 }
 
@@ -559,7 +576,6 @@ MGAPaletteLoadCallback(ScrnInfoPtr pScrn)
 	}
 	pal++;
     }
-    
     pMga->PaletteLoadCallback = NULL;
 }
 
@@ -575,7 +591,7 @@ void MGAGLoadPalette(
     if((pMga->CurrentLayout.Overlay8Plus24) && (pVisual->nplanes != 8)) 
 	return;
 
-     if(pMga->Chipset == PCI_CHIP_MGAG400){ 
+     if(pMga->Chipset == PCI_CHIP_MGAG400 || pMga->Chipset == PCI_CHIP_MGAG550){ 
 	 /* load them at the retrace in the block handler instead to 
 	    work around some problems with static on the screen */
 	while(numColors--) {
@@ -640,57 +656,84 @@ MGAGRestore(ScrnInfoPtr pScrn, vgaRegPtr vgaReg, MGARegPtr mgaReg,
 	MGAPtr pMga = MGAPTR(pScrn);
 	CARD32 optionMask;
 
-	MGA_NOT_HAL(
-	/*
-	 * Code is needed to get things back to bank zero.
-	 */
-	 
-	/* restore DAC registers 
-	 * according to the docs we shouldn't write to reserved regs*/
-	for (i = 0; i < DACREGSIZE; i++) {
-	    if( (i <= 0x03) ||
-	    	(i == 0x07) ||
-	    	(i == 0x0b) ||
-	    	(i == 0x0f) ||
-	       ((i >= 0x13) && (i <= 0x17)) ||
-	    	(i == 0x1b) ||
-	    	(i == 0x1c) ||
-	       ((i >= 0x1f) && (i <= 0x29)) ||
-	       ((i >= 0x30) && (i <= 0x37)) )
-	       		continue; 
-	    outMGAdac(i, mgaReg->DacRegs[i]);
-	}
+        if(!pMga->SecondCrtc) {
+MGA_NOT_HAL(
+	   /*
+	    * Code is needed to get things back to bank zero.
+	    */
+	   
+	   /* restore DAC registers 
+	    * according to the docs we shouldn't write to reserved regs*/
+	   for (i = 0; i < DACREGSIZE; i++) {
+	      if( (i <= 0x03) ||
+		  (i == 0x07) ||
+		  (i == 0x0b) ||
+		  (i == 0x0f) ||
+		  ((i >= 0x13) && (i <= 0x17)) ||
+		  (i == 0x1b) ||
+		  (i == 0x1c) ||
+		  ((i >= 0x1f) && (i <= 0x29)) ||
+		  ((i >= 0x30) && (i <= 0x37)) ||
+                  (MGAISGx50(pMga) && !mgaReg->PIXPLLCSaved &&
+		   ((i == 0x2c) || (i == 0x2d) || (i == 0x2e) ||
+		    (i == 0x4c) || (i == 0x4d) || (i == 0x4e))))
+		 continue; 
+	      outMGAdac(i, mgaReg->DacRegs[i]);
+	   }
+	   
+	   /* Do not set the memory config for primary cards as it
+	      should be correct already */
+	   optionMask = (pMga->Primary) ? OPTION1_MASK_PRIMARY : OPTION1_MASK; 
+	   
+	   if (!MGAISGx50(pMga)) {
+	      /* restore pci_option register */
+	      pciSetBitsLong(pMga->PciTag, PCI_OPTION_REG, optionMask,
+			     mgaReg->Option);
+	      if (pMga->Chipset != PCI_CHIP_MGA1064)
+		 pciSetBitsLong(pMga->PciTag, PCI_MGA_OPTION2, OPTION2_MASK,
+				mgaReg->Option2);
+	      if (pMga->Chipset == PCI_CHIP_MGAG400 || pMga->Chipset == PCI_CHIP_MGAG550)
+		 pciSetBitsLong(pMga->PciTag, PCI_MGA_OPTION3, OPTION3_MASK,
+				mgaReg->Option3);
+	   }
+);	/* MGA_NOT_HAL */
+	
+	   /* restore CRTCEXT regs */
+           for (i = 0; i < 6; i++)
+	      OUTREG16(0x1FDE, (mgaReg->ExtVga[i] << 8) | i);
 
-	/* Do not set the memory config for primary cards as it
-	   should be correct already */
-	optionMask = (pMga->Primary) ? OPTION1_MASK_PRIMARY : OPTION1_MASK; 
-
-	/* restore pci_option register */
-	pciSetBitsLong(pMga->PciTag, PCI_OPTION_REG, optionMask,
-		       mgaReg->Option);
-	if (pMga->Chipset != PCI_CHIP_MGA1064)
-		pciSetBitsLong(pMga->PciTag, PCI_MGA_OPTION2, OPTION2_MASK,
-			       mgaReg->Option2);
-	if (pMga->Chipset == PCI_CHIP_MGAG400)
-		pciSetBitsLong(pMga->PciTag, PCI_MGA_OPTION3, OPTION3_MASK,
-			       mgaReg->Option3);
-	);	/* MGA_NOT_HAL */
-
-	/* restore CRTCEXT regs */
-	for (i = 0; i < 6; i++)
-		OUTREG16(0x1FDE, (mgaReg->ExtVga[i] << 8) | i);
-
-	/*
-	 * This function handles restoring the generic VGA registers.
-	 */
-	vgaHWRestore(pScrn, vgaReg,
+	   /*
+	    * This function handles restoring the generic VGA registers.
+	    */
+	   vgaHWRestore(pScrn, vgaReg,
 			VGA_SR_MODE | (restoreFonts ? VGA_SR_FONTS : 0));
-	MGAGRestorePalette(pScrn, vgaReg->DAC);
+	   MGAGRestorePalette(pScrn, vgaReg->DAC);
+	   
+	   /*
+	    * this is needed to properly restore start address
+	    */
+	   OUTREG16(0x1FDE, (mgaReg->ExtVga[0] << 8) | 0);
+	} else {
+	   /* Second Crtc */
+	   xMODEINFO ModeInfo;
 
-	/*
-	 * this is needed to properly restore start address
-	 */
-	OUTREG16(0x1FDE, (mgaReg->ExtVga[0] << 8) | 0);
+MGA_NOT_HAL(
+	   /* Enable Dual Head */
+	   MGACRTC2Set(pScrn, &ModeInfo); 
+	   MGAEnableSecondOutPut(pScrn, &ModeInfo); 
+	   MGACRTC2SetPitch(pScrn, &ModeInfo); 
+	   MGACRTC2SetDisplayStart(pScrn, &ModeInfo,0,0,0);
+            
+	   for (i = 0x80; i <= 0xa0; i ++) {
+                if (i== 0x8d) {
+		   i = 0x8f;
+		   continue;
+		}
+                outMGAdac(i,   mgaReg->dac2[ i - 0x80]);
+	   }
+); /* MGA_NOT_HAL */
+
+        } 
 
 #ifdef DEBUG		
 	ErrorF("Setting DAC:");
@@ -723,7 +766,13 @@ MGAGSave(ScrnInfoPtr pScrn, vgaRegPtr vgaReg, MGARegPtr mgaReg,
 	int i;
 	MGAPtr pMga = MGAPTR(pScrn);
 
-	if(pMga->SecondCrtc == TRUE) return;
+	if(pMga->SecondCrtc == TRUE) {
+	   for(i = 0x80; i < 0xa0; i++)
+	      mgaReg->dac2[i-0x80] = inMGAdac(i);
+
+	   return;
+	}
+
 	MGA_NOT_HAL(
 	/* Allocate the DacRegs space if not done already */
 	if (mgaReg->DacRegs == NULL) {
@@ -750,10 +799,12 @@ MGAGSave(ScrnInfoPtr pScrn, vgaRegPtr vgaReg, MGARegPtr mgaReg,
 	for (i = 0; i < DACREGSIZE; i++)
 		mgaReg->DacRegs[i] = inMGAdac(i);
 
+        mgaReg->PIXPLLCSaved = TRUE;
+
 	mgaReg->Option = pciReadLong(pMga->PciTag, PCI_OPTION_REG);
 
 	mgaReg->Option2 = pciReadLong(pMga->PciTag, PCI_MGA_OPTION2);
-	if (pMga->Chipset == PCI_CHIP_MGAG400)
+	if (pMga->Chipset == PCI_CHIP_MGAG400 || pMga->Chipset == PCI_CHIP_MGAG550)
 	    mgaReg->Option3 = pciReadLong(pMga->PciTag, PCI_MGA_OPTION3);
 	);	/* MGA_NOT_HAL */
 
@@ -806,6 +857,14 @@ MGAGShowCursor(ScrnInfoPtr pScrn)
     outMGAdac(MGA1064_CURSOR_CTL, 0x03);
 }
 
+static void 
+MGAGShowCursorG100(ScrnInfoPtr pScrn)
+{
+    MGAPtr pMga = MGAPTR(pScrn);
+    /* Enable cursor - X-Windows mode */
+    outMGAdac(MGA1064_CURSOR_CTL, 0x01);
+}
+
 static void
 MGAGHideCursor(ScrnInfoPtr pScrn)
 {
@@ -825,9 +884,13 @@ MGAGSetCursorPosition(ScrnInfoPtr pScrn, int x, int y)
     while( INREG( MGAREG_Status ) & 0x08 );
     
     /* Output position - "only" 12 bits of location documented */
-    OUTREG( RAMDAC_OFFSET + MGA1064_CUR_XLOW, (y << 16) | (x & 0xFFFF));
+    OUTREG8( RAMDAC_OFFSET + MGA1064_CUR_XLOW, (x & 0xFF));
+    OUTREG8( RAMDAC_OFFSET + MGA1064_CUR_XHI, (x & 0xF00) >> 8);
+    OUTREG8( RAMDAC_OFFSET + MGA1064_CUR_YLOW, (y & 0xFF));
+    OUTREG8( RAMDAC_OFFSET + MGA1064_CUR_YHI, (y & 0xF00) >> 8);
 }
 
+
 static void
 MGAGSetCursorColors(ScrnInfoPtr pScrn, int bg, int fg)
 {
@@ -844,6 +907,22 @@ MGAGSetCursorColors(ScrnInfoPtr pScrn, int bg, int fg)
     outMGAdac(MGA1064_CURSOR_COL1_BLUE,  (fg & 0x000000FF));
 }
 
+static void
+MGAGSetCursorColorsG100(ScrnInfoPtr pScrn, int bg, int fg)
+{
+    MGAPtr pMga = MGAPTR(pScrn);
+
+    /* Background color */
+    outMGAdac(MGA1064_CURSOR_COL1_RED,   (bg & 0x00FF0000) >> 16);
+    outMGAdac(MGA1064_CURSOR_COL1_GREEN, (bg & 0x0000FF00) >> 8);
+    outMGAdac(MGA1064_CURSOR_COL1_BLUE,  (bg & 0x000000FF));
+
+    /* Foreground color */
+    outMGAdac(MGA1064_CURSOR_COL2_RED,   (fg & 0x00FF0000) >> 16);
+    outMGAdac(MGA1064_CURSOR_COL2_GREEN, (fg & 0x0000FF00) >> 8);
+    outMGAdac(MGA1064_CURSOR_COL2_BLUE,  (fg & 0x000000FF));
+}
+
 static Bool 
 MGAGUseHWCursor(ScreenPtr pScrn, CursorPtr pCurs)
 {
@@ -951,6 +1030,7 @@ MGAG_i2cInit(ScrnInfoPtr pScrn)
 
 /*
  * MGAGRamdacInit
+ * Handle broken G100 special.
  */
 static void
 MGAGRamdacInit(ScrnInfoPtr pScrn)
@@ -962,14 +1042,22 @@ MGAGRamdacInit(ScrnInfoPtr pScrn)
     MGAdac->CursorOffscreenMemSize = 1024;
     MGAdac->CursorMaxWidth         = 64;
     MGAdac->CursorMaxHeight        = 64;
-    MGAdac->SetCursorColors        = MGAGSetCursorColors;
     MGAdac->SetCursorPosition      = MGAGSetCursorPosition;
     MGAdac->LoadCursorImage        = MGAGLoadCursorImage;
     MGAdac->HideCursor             = MGAGHideCursor;
-    MGAdac->ShowCursor             = MGAGShowCursor;
+    if ((pMga->Chipset == PCI_CHIP_MGAG100) 
+	|| (pMga->Chipset == PCI_CHIP_MGAG100)) {
+      MGAdac->SetCursorColors        = MGAGSetCursorColorsG100;
+      MGAdac->ShowCursor             = MGAGShowCursorG100;
+    } else {
+      MGAdac->SetCursorColors        = MGAGSetCursorColors;
+      MGAdac->ShowCursor             = MGAGShowCursor;
+    }
     MGAdac->UseHWCursor            = MGAGUseHWCursor;
     MGAdac->CursorFlags            =
+#if X_BYTE_ORDER == X_LITTLE_ENDIAN
     				HARDWARE_CURSOR_BIT_ORDER_MSBFIRST |
+#endif
     				HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_64 |
     				HARDWARE_CURSOR_TRUECOLOR_AT_8BPP;
 
@@ -991,6 +1079,7 @@ MGAGRamdacInit(ScrnInfoPtr pScrn)
 	        MGAdac->maxPixelClock = 220000;
 	    break;
     	case PCI_CHIP_MGAG400:
+    	case PCI_CHIP_MGAG550:
 	    /* We don't know the new pins format but we know that
 	       the maxclock / 4 is where the RamdacType was in the
 	       old pins format */
